from __future__ import annotations as _annotations

# ==( Custom Annotations )==
# This file contains private annotations used across this package.
# Whenever there is a "?" comment, it means that
# a type may or may not implement that field or mixin class.

from typing import (
    TypeVar as _TypeVar,
    Protocol as _Protocol,
    ClassVar as _ClassVar,
    Any as _Any,
    TYPE_CHECKING as _TYPE_CHECKING,
)

from linflex import (
    Vec2 as _Vec2,
    Vec2i as _Vec2i,
)
from colex import ColorValue as _ColorValue

if _TYPE_CHECKING:
    from ._clock import DeltaClock as _DeltaClock
    from ._screen import Screen as _Screen
    from ._animation import (
        Animation as _Animation,
        AnimationMapping as _AnimationMapping,
    )

EngineType = _TypeVar("EngineType", bound="Engine", covariant=True)
NodeType = _TypeVar("NodeType", bound="Node", covariant=True)
T = _TypeVar("T")
_Self = _TypeVar("_Self")
_T_contra = _TypeVar("_T_contra", contravariant=True)


class FileLike(_Protocol[_T_contra]):
    def write(self, stream: _T_contra, /) -> object: ...
    def flush(self, /) -> None: ...
    def fileno(self, /) -> int: ...


class Engine(_Protocol):
    fps: float
    clock: _DeltaClock
    screen: _Screen
    is_running: bool


class Node(_Protocol):
    node_instances: _ClassVar[dict[int, Node]]
    uid: int

    def __init__(self) -> None: ...
    def with_parent(self: _Self, parent: Node | None, /) -> _Self: ...
    def with_process_priority(self, process_priority: int, /): ...
    def update(self, delta: float) -> None: ...
    def queue_free(self) -> None: ...
    def free(self) -> None: ...


class TransformComponent(_Protocol):
    transform_instances: _ClassVar[dict[int, TransformNode]]
    position: _Vec2
    rotation: float
    z_index: int
    is_top_level: bool

    def with_position(
        self: _Self,
        position: _Vec2 | None = None,
        /,
        x: float | None = None,
        y: float | None = None,
    ) -> _Self: ...
    def with_global_position(
        self: _Self,
        global_position: _Vec2 | None = None,
        /,
        x: float | None = None,
        y: float | None = None,
    ) -> _Self: ...
    def with_rotation(self: _Self, rotation: float, /) -> _Self: ...
    def with_global_rotation(self: _Self, global_rotation: float, /) -> _Self: ...
    def with_z_index(self: _Self, z_index: int, /) -> _Self: ...
    def as_top_level(self: _Self, state: bool = True, /) -> _Self: ...
    @property
    def global_position(self) -> _Vec2: ...
    @global_position.setter
    def global_position(self, position: _Vec2) -> None: ...
    @property
    def global_rotation(self) -> float: ...
    @global_rotation.setter
    def global_rotation(self, rotation: float) -> None: ...


class TransformNode(
    TransformComponent,
    Node,
    _Protocol,
): ...


class TextureComponent(_Protocol):
    texture_instances: _ClassVar[dict[int, TextureNode]]
    texture: list[str]
    visible: bool
    centered: bool
    transparency: str | None

    def with_texture(self: _Self, texture_or_line: list[str] | str, /) -> _Self: ...
    def as_visible(self: _Self, state: bool = True, /) -> _Self: ...
    def as_centered(self: _Self, state: bool = True, /) -> _Self: ...
    def with_transparency(self: _Self, state: bool = True, /) -> _Self: ...
    def hide(self) -> None: ...
    def show(self) -> None: ...
    def is_globally_visible(self) -> bool: ...
    def get_texture_size(self) -> _Vec2i: ...


class TextureNode(
    TextureComponent,
    TransformComponent,
    Node,
    _Protocol,
): ...


class ColorComponent(_Protocol):
    color_instances: _ClassVar[dict[int, ColorNode]]
    color: _ColorValue | None

    def with_color(
        self: _Self,
        color: _ColorValue,
        /,
    ) -> _Self: ...


class ColorNode(
    ColorComponent,
    TextureComponent,
    TransformNode,
    Node,
    _Protocol,
): ...


class Renderable(  # possible base: `ColorComponent`
    TextureComponent,
    TransformComponent,
    Node,
    _Protocol,
):
    ...
    # possible field: `color` of type `_ColorValue`


class AnimatedComponent(_Protocol):
    animated_instances: _ClassVar[dict[int, AnimatedNode]]
    animations: _AnimationMapping
    current_animation: _Animation | None = None
    _frame_index: int = 0

    def with_animations(self: _Self, animations: dict[str, _Any], /) -> _Self: ...
    def with_animation(
        self: _Self,
        animation_name: str,
        animation: _Animation,
        /,
    ) -> _Self: ...
    def add_animation(
        self,
        animation_name: str,
        animation: _Animation,
    ) -> None: ...
    def play(self, animation_name: str, /) -> None: ...
    def _wrapped_update_animated(self, _delta: float) -> None: ...


class AnimatedNode(  # possible base: `ColorComponent`
    AnimatedComponent,
    TextureComponent,
    TransformComponent,
    Node,
    _Protocol,
):
    ...
    # possible field: `color` of type `_ColorValue`



from __future__ import annotations as _annotations

from typing import (
    Generator as _Generator,
    Any as _Any,
)

from colex import ColorValue as _ColorValue

from ._annotations import (
    NodeType as _NodeType,
    ColorNode as _ColorNode,
)


class Color:
    color_instances: dict[int, _ColorNode] = {}

    def __new__(cls: type[_NodeType], *args: _Any, **kwargs: _Any) -> _NodeType:
        instance = super().__new__(cls, *args, **kwargs)  # type: _ColorNode  # type: ignore[reportAssignmentType]
        Color.color_instances[instance.uid] = instance
        return instance  # type: ignore

    color: _ColorValue | None = None

    def with_color(self, color: _ColorValue | None, /):
        self.color = color
        return self

    def free(self: _ColorNode) -> None:
        del Color.color_instances[self.uid]
        super().free()



from __future__ import annotations as _annotations

from copy import deepcopy as _deepcopy
from typing import (
    Generator as _Generator,
    Any as _Any,
    ClassVar as _ClassVar,
)

from linflex import Vec2 as _Vec2

from ._annotations import (
    NodeType as _NodeType,
    TransformNode as _TransformNode,
)


class Transform:
    transform_instances: _ClassVar[dict[int, _TransformNode]] = {}

    def __new__(cls: type[_NodeType], *args: _Any, **kwargs: _Any) -> _NodeType:
        instance = super().__new__(cls, *args, **kwargs)  # type: _TransformNode  # type: ignore[reportAssignmentType]
        Transform.transform_instances[instance.uid] = instance
        if (class_position := getattr(instance, "position", None)) is not None:
            instance.position = _deepcopy(class_position)
        else:
            instance.position = _Vec2.ZERO
        return instance  # type: ignore

    position: _Vec2
    rotation: float = 0
    z_index: int = 0
    is_top_level: bool = False

    # TODO: would be nice to figure out @overload with this function
    def with_position(
        self,
        position: _Vec2 | None = None,
        /,
        x: float | None = None,
        y: float | None = None,
    ):
        if position is None and x is None and y is None:
            raise TypeError(f"not all arguments can be {None} at the same time")
        if position is not None and (x is not None or y is not None):
            raise TypeError(
                "chose either positional argument 'position' "
                "or keyword arguments 'x' and/or 'y', not all three"
            )
        if position is not None:
            self.position = position
        if x is not None or y is not None:
            self.position = _Vec2(x or 0, y or 0)
        return self

    # TODO: would be nice to figure out @overload with this function
    def with_global_position(
        self,
        global_position: _Vec2 | None = None,
        /,
        x: float | None = None,
        y: float | None = None,
    ):
        if global_position is None and x is None and y is None:
            raise TypeError(f"not all arguments can be {None} at the same time")
        if global_position is not None and (x is not None or y is not None):
            raise TypeError(
                "chose either positional argument 'global_position' "
                "or keyword arguments 'x' and/or 'y', not all three"
            )
        if global_position is not None:
            self.global_position = global_position
        if x is not None or y is not None:
            self.global_position = _Vec2(x or 0, y or 0)
        return self

    def with_rotation(self, rotation: float, /):
        self.rotation = rotation
        return self

    def with_global_rotation(self, global_rotation: float, /):
        self.global_rotation = global_rotation
        return self

    def with_z_index(self, z_index: int, /):
        self.z_index = z_index
        return self

    def as_top_level(self, state: bool = True, /):
        self.is_top_level = state
        return self

    @property
    def global_position(self) -> _Vec2:
        """Computes the node's global position (world space)

        Returns:
            _Vec2: global position
        """
        if self.is_top_level:
            return self.position.copy()
        global_position = self.position.copy()
        parent = self.parent  # type: ignore
        while parent is not None and isinstance(parent, Transform):
            global_position = parent.position + global_position.rotated(parent.rotation)
            if parent.is_top_level:
                return global_position
            parent = parent.parent  # type: ignore
        return global_position

    @global_position.setter
    def global_position(self, position: _Vec2) -> None:
        """Sets the node's global position (world space)"""
        diff = position - self.global_position
        self.position += diff

    @property
    def global_rotation(self) -> float:
        """Computes the node's global rotation (world space)

        Returns:
            float: global rotation in radians
        """
        if self.is_top_level:
            return self.rotation
        global_rotation = self.rotation
        parent = self.parent  # type: ignore
        while parent is not None and isinstance(parent, Transform):
            global_rotation += parent.rotation
            if parent.is_top_level:
                return global_rotation
            parent = parent.parent  # type: ignore
        return global_rotation

    @global_rotation.setter
    def global_rotation(self, rotation: float) -> None:
        """Sets the node's global rotation (world space)"""
        diff = rotation - self.global_rotation
        self.rotation += diff

    def free(self: _TransformNode) -> None:
        del Transform.transform_instances[self.uid]
        super().free()



"""
Charz
-----

An object oriented terminal game engine

Includes:
- `Engine`
- `Clock`
- `DeltaClock`
- `Screen`
- `Camera`
- `CameraMode` (enum)
- `Node` (base node)
- `Node2D`
- `Transform` (component)
- `Vec2` (datastructure from `linflex` package)
- `Texture` (component)
- `Color` (component)
- `ColorValue` (annotation from `colex` package)
- `Label` (prefabricated)
- `Sprite` (prefabricated)
- `AnimatedSprite` (prefabricated)
- `Animation` (datastructure)
- `text` (module for flipping strings)
"""

from __future__ import annotations as _annotations

__version__ = "0.0.17"
__all__ = [
    "Engine",
    "Clock",
    "DeltaClock",
    "Screen",
    "Camera",
    "CameraMode",
    "Node",
    "Node2D",
    "Transform",
    "Vec2",
    "load_texture",
    "Texture",
    "Color",
    "ColorValue",
    "Label",
    "Sprite",
    "AnimatedSprite",
    "Animation",
    "text",
]

from linflex import Vec2
from colex import ColorValue

from ._engine import Engine
from ._clock import Clock, DeltaClock
from ._screen import Screen
from ._camera import Camera, CameraMode
from ._node import Node
from ._transform import Transform
from ._texture import load_texture, Texture
from ._color import Color
from ._animation import Animation
from ._prefabs._node2d import Node2D
from ._prefabs._label import Label
from ._prefabs._sprite import Sprite
from ._prefabs._animated_sprite import AnimatedSprite
from . import text



from __future__ import annotations as _annotations

from types import SimpleNamespace as _SimpleNamespace
from functools import wraps as _wraps
from pathlib import Path as _Path
from copy import deepcopy as _deepcopy
from typing import (
    Generator as _Generator,
    Any as _Any,
    ClassVar as _ClassVar,
)

from ._texture import load_texture as _load_texture
from ._annotations import (
    T as _T,
    NodeType as _NodeType,
    AnimatedNode as _AnimatedNode,
)


class Animation:
    __slots__ = ("frames",)
    frames: list[list[str]]

    def __init__(self, animation_path: _Path | str, /) -> None:
        # fmt: off
        frame_directory = (
            _Path.cwd()
            .joinpath(str(animation_path))
            .iterdir()
        )
        # fmt: on
        self.frames = list(map(_load_texture, frame_directory))


class AnimationMapping(_SimpleNamespace):
    def __init__(self, **animations: Animation) -> None:
        super().__init__(**animations)

    def __getattribute__(self, name: str) -> Animation:
        return super().__getattribute__(name)

    def __setattr__(self, name: str, value: Animation) -> None:
        return super().__setattr__(name, value)

    def get(self, animation_name: str, default: _T = None) -> Animation | _T:
        return getattr(self, animation_name, default)

    def update(self, animations: dict[str, Animation]) -> None:
        for name, animation in animations.items():
            setattr(self, name, animation)


# TODO: add `.play_backwards` attribute or method
class Animated:  # Component (mixin class)
    animated_instances: _ClassVar[dict[int, _AnimatedNode]] = {}

    def __new__(cls: type[_NodeType], *args: _Any, **kwargs: _Any) -> _NodeType:
        instance = super().__new__(cls, *args, **kwargs)  # type: _AnimatedNode  # type: ignore[reportAssignmentType]
        Animated.animated_instances[instance.uid] = instance
        if (class_animations := getattr(instance, "animations", None)) is not None:
            instance.animations = _deepcopy(class_animations)
        else:
            instance.animations = AnimationMapping()

        # inject `._wrapped_update_animated()` into `.update()`
        def update_method_factory(instance: _AnimatedNode, bound_update):  # noqa: ANN001 ANN202
            @_wraps(bound_update)
            def new_update_method(delta: float) -> None:
                bound_update(delta)
                instance._wrapped_update_animated(delta)

            return new_update_method

        instance.update = update_method_factory(instance, instance.update)
        return instance  # type: ignore

    animations: AnimationMapping
    current_animation: Animation | None = None
    is_playing: bool = False
    _frame_index: int = 0

    def with_animations(self, /, **animations: Animation):
        self.animations.update(animations)
        return self

    def with_animation(
        self,
        animation_name: str,
        animation: Animation,
        /,
    ):
        self.add_animation(animation_name, animation)
        return self

    def add_animation(
        self,
        animation_name: str,
        animation: Animation,
        /,
    ) -> None:
        setattr(self.animations, animation_name, animation)

    def play(self, animation_name: str, /) -> None:
        self.current_animation = self.animations.get(animation_name, None)
        self.is_playing = True
        self._frame_index = 0
        # the actual logic of playing the animation is handled in `.update(...)`

    def _wrapped_update_animated(self, _delta: float) -> None:
        if self.current_animation is None:
            self.is_playing = False
            return
        self.texture = self.current_animation.frames[self._frame_index]
        frame_count = len(self.current_animation.frames)
        self._frame_index = min(self._frame_index + 1, frame_count - 1)
        if self._frame_index == frame_count - 1:
            self.is_playing = False

    def free(self: _AnimatedNode) -> None:
        del Animated.animated_instances[self.uid]
        super().free()



from __future__ import annotations as _annotations

from enum import (
    IntEnum as _IntEnum,
    auto as _auto,
)
from typing import ClassVar as _ClassVar

from ._node import Node as _Node
from ._transform import Transform as _Transform


class CameraMode(_IntEnum):
    FIXED = 0
    CENTERED = _auto()
    INCLUDE_SIZE = _auto()

    def __or__(self, other: CameraMode) -> CameraMode:
        # NOTE: result may not be a member of 'CameraMode', but instead an 'int'.
        # combine any 'CameraMode', where any value that is not bound to a variant,
        # will in fact be an 'int', but we treat it like a 'CameraMode' that was combined
        if isinstance(other, CameraMode):
            new_value = self.value | other.value
            if new_value in CameraMode._value2member_map_:
                return CameraMode(new_value)
            return new_value  # type: ignore
        return NotImplemented


class Camera(_Transform, _Node):
    current: _ClassVar[Camera]
    mode: CameraMode = CameraMode.FIXED

    def set_current(self) -> None:
        Camera.current = self

    def as_current(self, state: bool = True):
        if state:
            self.set_current()
            return self
        Camera.current = Camera()  # make new default camera
        Camera.current.free()  # remove from node count, will still be used as placeholder
        return self

    def is_current(self) -> bool:
        return Camera.current is self

    def with_mode(self, mode: CameraMode, /):
        self.mode = mode
        return self


Camera.current = Camera()  # initial camera
Camera.current.free()  # remove from node count, will still be used as placeholder



from __future__ import annotations as _annotations

import time as _time


class Clock:
    """`Clock` base class, without delta time"""

    def __init__(self, tps: float = 16, /) -> None:
        """Initializes the clock with a given tps

        Args:
            tps (float, optional): ticks per second. Defaults to 16.
        """
        self.tps = tps
        self._delta_time = 1.0 / tps  # average delta time

    def with_tps(self, tps: float):
        self.tps = tps
        return self

    @property
    def tps(self) -> float:
        return self._tps

    @tps.setter
    def tps(self, value: float) -> None:
        self._tps = value

    def get_delta(self) -> float:
        """Gets the time it took until tick was called again

        Returns:
            float: delta time until last fram
        """
        return self._delta_time

    def tick(self) -> None:
        """Does nothing. Exists for better coupling, when extending the `Clock` class"""
        return


class DeltaClock(Clock):
    """`DeltaClock` calculating `delta time` and sleeps for maintaining desirerd `tps`"""

    def __init__(self, tps: float = 16, /) -> None:
        """Initializes the delta clock with a given tps

        Args:
            tps (float, optional): ticks per second. Defaults to 16.
        """
        self.tps = tps  # calls property setter
        self._delta_time = 1.0 / tps  # initial delta time (optimal scenario)
        self._last_tick = _time.perf_counter()

    @property
    def tps(self) -> float:
        return self._tps

    @tps.setter
    def tps(self, value: float) -> None:
        self._tps = value
        self._target_delta = 1.0 / self._tps

    def tick(self) -> None:
        """Pauses the clock temporay to achieve the desired framerate (tps)"""
        current_time = _time.perf_counter()
        elapsed_time = current_time - self._last_tick
        sleep_time = self._target_delta - elapsed_time
        if sleep_time > 0:
            _time.sleep(sleep_time)
            self._last_tick = _time.perf_counter()
        else:
            self._last_tick = current_time
        self._delta_time = max(0, sleep_time)



from __future__ import annotations as _annotations

import os as _os
import sys as _sys

from linflex import Vec2i as _Vec2i
from colex import (
    ColorValue as _ColorValue,
    RESET as _RESET,
)

from ._camera import (
    Camera as _Camera,
    CameraMode as _CameraMode,
)
from ._transform import Transform as _Transform
from ._texture import Texture as _Texture
from ._annotations import (
    FileLike as _FileLike,
    Renderable as _Renderable,
)


class Screen:
    stream: _FileLike[str] = _sys.stdout  # default stream, may be redirected
    # screen texture buffer with (char, color) tuple
    buffer: list[list[tuple[str, _ColorValue | None]]]

    def __init__(
        self,
        width: int = 16,
        height: int = 12,
        *,
        auto_resize: bool = False,
        transparancy_fill: str = " ",
        background_color: _ColorValue | None = None,
        margin_right: int = 1,
        margin_bottom: int = 1,
    ) -> None:
        self.width = width
        self.height = height
        self.margin_right = margin_right
        self.margin_bottom = margin_bottom
        self._auto_resize = auto_resize
        self._resize_if_necessary()
        self.transparancy_fill = transparancy_fill
        self.background_color = background_color
        self.buffer = []
        self.clear()  # for populating the list with an empty screen

    @property
    def auto_resize(self) -> bool:
        return self._auto_resize

    @auto_resize.setter
    def auto_resize(self, state: bool) -> None:
        self._auto_resize = state
        self._resize_if_necessary()

    def _resize_if_necessary(self) -> None:  # NOTE: does not mutate screen buffer
        if self.auto_resize:
            terminal_size = _os.get_terminal_size(self.stream.fileno())
            self.width = terminal_size.columns - self.margin_right
            self.height = terminal_size.lines - self.margin_bottom

    def with_stream(self, stream: _FileLike[str], /):
        self.stream = stream
        return self

    def with_width(self, width: int, /):
        self.width = width
        return self

    def with_height(self, height: int, /):
        self.height = height
        return self

    def with_size(self, size: _Vec2i, /):
        self.size = size
        return self

    def with_auto_resize(self, state: bool = True, /):
        self.auto_resize = state
        return self

    def with_transparancy_fill(self, fill_char: str, /):
        if len(fill_char) != 1:
            raise ValueError(
                "'transparancy_fill' cannot be a 'str' with length other than 1,"
                f"got length {len(fill_char)}"
            )
        self.transparancy_fill = fill_char
        return self

    @property
    def size(self) -> _Vec2i:
        return _Vec2i(self.width, self.height)

    @size.setter
    def size(self, value: _Vec2i) -> None:
        width, height = value.to_tuple()
        if not isinstance(width, int):
            raise ValueError(f"width cannot be of type '{type(value)}', expected 'int'")
        if not isinstance(height, int):
            raise ValueError(f"height cannot be of type '{type(value)}', expected 'int'")
        self.width = width
        self.height = height
        self._resize_if_necessary()

    def clear(self) -> None:
        self.buffer = [
            # (char, color) group
            [(self.transparancy_fill, self.background_color) for _ in range(self.width)]
            for _ in range(self.height)
        ]

    def render(self, node: _Renderable, /) -> None:  # noqa: C901
        if not node.is_globally_visible():  # skip if node is invisible
            return
        # current camera should never be None or other class than 'Camera',
        # or subclass of it
        if _Camera.current is None or not isinstance(_Camera.current, _Camera):
            raise TypeError(
                "'Camera.current' cannot be of type "
                f"'{type(_Camera.current)}' while rendering"
            )

        color: _ColorValue | None = getattr(node, "color")  # noqa: B009
        # TODO: implement rotation when rendering
        # node_global_rotation = node.global_rotation
        node_global_position = node.global_position

        # determine whether to use use the parent of current camera
        # or its parent as anchor for viewport
        anchor = _Camera.current
        if _Camera.current.parent is not None and isinstance(
            _Camera.current.parent, _Transform
        ):
            anchor = _Camera.current.parent
        relative_position = node_global_position - anchor.global_position

        if _Camera.current.mode & _CameraMode.CENTERED:
            relative_position += self.size / 2

        # include half size of camera parent when including size
        viewport_global_position = _Camera.current.global_position
        if (
            _Camera.current.mode & _CameraMode.INCLUDE_SIZE
            and _Camera.current.parent is not None
            and isinstance(_Camera.current.parent, _Texture)
        ):
            # adds half of camera's parent's texture size
            viewport_global_position += _Camera.current.parent.get_texture_size() / 2

        terminal_size = _os.get_terminal_size()
        actual_width = min(self.width, terminal_size.columns - self.margin_right)
        actual_height = min(self.height, terminal_size.lines - self.margin_bottom)

        texture_size = node.get_texture_size()
        x = int(relative_position.x)
        y = int(relative_position.y)
        if node.centered:
            x = int(relative_position.x - (texture_size.x / 2))
            y = int(relative_position.y - (texture_size.y / 2))

        # out of bounds
        # TODO: consider nodes with rotation
        if x + texture_size.x < 0 or x > actual_width:
            return
        if y + texture_size.y < 0 or y > actual_height:
            return

        for y_offset, line in enumerate(node.texture):
            y_final = y + y_offset
            for x_offset, char in enumerate(line):
                if char == node.transparency:  # skip transparent char
                    continue
                x_final = x + x_offset
                # insert char into screen buffer if visible
                if 0 <= x_final < actual_width and 0 <= y_final < actual_height:
                    self.buffer[y_final][x_final] = (char, color)
        # TODO: implement render with rotation

    def show(self) -> None:
        size = _os.get_terminal_size()
        actual_width = min(self.width, size.columns - self.margin_right)  # -1 is margin
        actual_height = min(self.height, size.lines - self.margin_bottom)
        out = ""
        # construct frame
        for lino, row in enumerate(self.buffer[:actual_height], start=1):
            for char, color in row[:actual_width]:
                if color is not None:
                    out += color + char
                else:
                    out += _RESET + char
            if lino != len(self.buffer):  # not at end
                out += "\n"
        out += _RESET
        # move cursor
        move_code = f"\x1b[{actual_height - 1}A" + "\r"
        out += move_code
        # write and flush
        self.stream.write(out)
        self.stream.flush()

    def refresh(self) -> None:
        self._resize_if_necessary()
        self.clear()
        for node in sorted(  # NOTE: iterator becomes a 'list'
            # NOTE: 'list' is faster than 'tuple', when copying
            list(_Texture.texture_instances.values()),  # iterate a copy
            key=lambda node: node.z_index,
        ):
            self.render(node)
        self.show()



from __future__ import annotations as _annotations

from copy import deepcopy as _deepcopy
from typing import (
    Callable as _Callable,  # noqa: F401
    Any as _Any,
)

from linflex import Vec2i as _Vec2i

from ._clock import (
    Clock as _Clock,
    DeltaClock as _DeltaClock,
)
from ._screen import Screen as _Screen
from ._node import Node as _Node
from ._annotations import EngineType as _EngineType


class _EngineMixinSortMeta(type):
    """Engine metaclass for initializing `Engine` subclass after other `mixin` classes"""

    def __new__(cls, name: str, bases: tuple[type, ...], attrs: dict[str, object]):
        def sorter(base: type) -> bool:
            return isinstance(base, Engine)

        sorted_bases = tuple(sorted(bases, key=sorter))
        new_type = super().__new__(cls, name, sorted_bases, attrs)
        return new_type


class Engine(metaclass=_EngineMixinSortMeta):
    fps: float = 16
    clock: _Clock = _DeltaClock()
    screen: _Screen = _Screen()
    clear_console: bool = False
    hide_cursor: bool = True
    is_running: bool = False

    def __new__(cls: type[_EngineType], *args: _Any, **kwargs: _Any) -> _EngineType:
        instance = super().__new__(cls, *args, **kwargs)  # type: _EngineType  # type: ignore[reportAssignmentType]
        # overrides `.clock.tps` with `.fps` set from class attribute
        instance.clock.tps = instance.fps
        return instance  # type: ignore

    def with_fps(self, fps: float, /):
        self.fps = fps
        self.clock.tps = fps
        return self

    def with_clock(self, clock: _Clock, /):
        self.clock = clock.with_tps(self.fps)
        return self

    def with_screen(self, screen: _Screen, /):
        self.screen = screen
        return self

    # TODO: Try implementing @overload for this method
    def with_screen_size(
        self,
        size: _Vec2i | None = None,
        /,
        width: int | None = None,
        height: int | None = None,
    ):
        if size is None and width is None and height is None:
            raise TypeError(f"not all arguments can be 'None' at the same time")
        if size is not None and (width is not None or height is not None):
            raise TypeError(
                "chose either positional argument 'size' "
                "or keyword arguments 'width' and/or 'height', not all three"
            )
        if size is not None:
            self.screen.size = size
        if width is not None or height is not None:
            self.screen.size = _Vec2i(width or 16, height or 12)
        return self

    def with_auto_resize_screen(self, state: bool = True, /):
        self.screen.auto_resize = state
        return self

    def update(self, delta: float) -> None: ...

    def run(self):
        # check if console/stream should be cleared
        if self.clear_console:
            clear_code = "\x1b[0J"
            self.screen.stream.write(clear_code)
            self.screen.stream.flush()
        # hide cursor
        if self.hide_cursor:
            hide_code = "\x1b[?25l"
            self.screen.stream.write(hide_code)
            self.screen.stream.flush()
        delta = self.clock.get_delta()
        self.is_running = True
        while self.is_running:
            self.update(delta)
            for queued_node in _Node._queued_nodes:
                queued_node.free()
            _Node._queued_nodes *= 0  # NOTE: faster way to do `.clear()`
            # NOTE: 'list' is faster than 'tuple', when copying
            for node in list(_Node.node_instances.values()):  # iterating copy
                node.update(delta)
            self.screen.refresh()
            self.clock.tick()
            delta = self.clock.get_delta()
        # show cursor if hidden
        if self.hide_cursor:
            hide_code = "\x1b[?25h"
            self.screen.stream.write(hide_code)
            self.screen.stream.flush()

        return self  # this is for convenience if the desire to read the app state arise



from __future__ import annotations as _annotations

from .._node import Node as _Node
from .._transform import Transform as _Transform
from .._texture import Texture as _Texture
from .._color import Color as _Color


class Label(_Color, _Texture, _Transform, _Node):
    newline: str = "\n"
    tab_size: int = 4
    tab_char: str = "\t"
    tab_fill: str = " "

    def with_newline(self, newline: str, /):
        self.newline = newline
        return self

    def with_tab_size(self, tab_size: int, /):
        self.tab_size = tab_size
        return self

    def with_tab_char(self, tab_char: str, /):
        self.tab_char = tab_char
        return self

    def with_tab_fill(self, tab_fill: str, /):
        self.tab_fill = tab_fill
        return self

    def with_text(self, text: str, /):
        self.text = text
        return self

    @property
    def text(self) -> str:
        joined_lines = self.newline.join(self.texture)
        return joined_lines.replace(self.tab_fill * self.tab_size, self.tab_char)

    @text.setter
    def text(self, value: str) -> None:
        tab_replaced = self.newline.replace(self.tab_char, self.tab_fill * self.tab_size)
        self.texture = value.split(tab_replaced)



from __future__ import annotations as _annotations

from .._node import Node as _Node
from .._transform import Transform as _Transform


class Node2D(_Transform, _Node):
    def __str__(self) -> str:
        return (
            self.__class__.__name__
            + "("
            + f"#{self.uid}"
            + f":{self.position}"
            + f":{self.rotation}"
            + ")"
        )



from __future__ import annotations as _annotations

from pathlib import Path as _Path
from copy import deepcopy as _deepcopy
from typing import (
    Generator as _Generator,
    Any as _Any,
    ClassVar as _ClassVar,
)

from linflex import Vec2i as _Vec2i

from ._annotations import (
    NodeType as _NodeType,
    TextureNode as _TextureNode,
)


def load_texture(file_path: _Path | str, /) -> list[str]:
    return _Path.cwd().joinpath(str(file_path)).read_text(encoding="utf-8").splitlines()


class Texture:
    texture_instances: _ClassVar[dict[int, _TextureNode]] = {}

    def __new__(cls: type[_NodeType], *args: _Any, **kwargs: _Any) -> _NodeType:
        instance = super().__new__(cls, *args, **kwargs)  # type: _TextureNode  # type: ignore[reportAssignmentType]
        Texture.texture_instances[instance.uid] = instance
        if (class_texture := getattr(instance, "texture", None)) is not None:
            instance.texture = _deepcopy(class_texture)
        else:
            instance.texture = []
        return instance  # type: ignore

    texture: list[str]
    visible: bool = True
    centered: bool = False
    transparency: str | None = None

    def with_texture(self, texture_or_line: list[str] | str, /):
        if isinstance(texture_or_line, str):
            self.texture = [texture_or_line]
            return self
        self.texture = texture_or_line
        return self

    def as_visible(self, state: bool = True, /):
        self.visible = state
        return self

    def as_centered(self, state: bool = True, /):
        self.centered = state
        return self

    def with_transparency(self, char: str | None, /):
        self.transparancy = char
        return self

    def hide(self) -> None:
        self.visible = False

    def show(self) -> None:
        self.visible = True

    def is_globally_visible(self) -> bool:  # global visibility
        """Checks whether the node and its ancestors are visible

        Returns:
            bool: global visibility
        """
        if not self.visible:
            return False
        parent = self.parent  # type: ignore
        while parent is not None:
            if not isinstance(parent, Texture):
                return True
            if not parent.visible:
                return False
            parent = parent.parent  # type: ignore
        return True

    def get_texture_size(self) -> _Vec2i:
        if not self.texture:
            return _Vec2i.ZERO
        return _Vec2i(
            len(max(self.texture, key=len)),  # size of longest line
            len(self.texture),  # line count
        )

    def free(self: _TextureNode) -> None:
        del Texture.texture_instances[self.uid]
        super().free()



from __future__ import annotations as _annotations

from .._node import Node as _Node
from .._transform import Transform as _Transform
from .._texture import Texture as _Texture
from .._color import Color as _Color


class Sprite(_Color, _Texture, _Transform, _Node):
    def __str__(self) -> str:
        return (
            self.__class__.__name__
            + "("
            + f"#{self.uid}"
            + f":{self.position}"
            + f":{round(self.rotation, 2)}R"
            + f":{'{}x{}'.format(*self.get_texture_size().to_tuple())}"
            + f":{repr(self.color)}"
            + ")"
        )
